from typing import Dict, List, Optional, TYPE_CHECKING
from .repo import Repository
from .utils import save_get

if TYPE_CHECKING:
    from .person import Person


class Team:
    id: 'int | str'
    key: str
    assignment: str
    persons: 'List[Person]'
    repository: Optional[Repository]

    def __init__(
        self, id_: 'int | str', assignment: str, persons: 'Optional[List[Person]]' = None,
        repository: Optional[Repository] = None
    ) -> None:
        self.id = id_
        self.assignment = assignment
        self.key = f"{assignment}-{id_}"
        self.persons = persons or []
        self.repository = repository

    @classmethod
    def from_csv_line(
        cls, headers: Dict[str, int], data: List[str], assignments: List[str], person: 'Person'
    ) -> 'List[Team]':
        ret: 'List[Team]' = []
        for ass in assignments:
            if not save_get(data, headers[f"Team{ass}"]):
                continue
            try:
                team = Team(
                    int(data[headers[f"Team{ass}"]]),
                    ass,
                    [person]
                )
            except ValueError:
                team = Team(
                    data[headers[f"Team{ass}"]],
                    ass,
                    [person]
                )
            if f"WebUrlRepo{ass}" in headers and data[headers[f"WebUrlRepo{ass}"]]:
                server = data[headers[f"WebUrlRepo{ass}"]].split("//", 1)[1].split("/", 1)
                team.repository = Repository(
                    server[0],
                    server[1],
                    int(data[headers[f"Repo{ass}ID"]])
                )
            ret.append(team)
        return ret

    def __repr__(self) -> str:
        return f"Team({self.id}, {self.assignment}, {self.persons!r}, {self.repository!r})"



from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Any, List, Dict, Literal, Iterable, Tuple, TYPE_CHECKING
import requests
import time

if TYPE_CHECKING:
    from .project import Project
    from .group import Group
    from .user import User
    from esprepomanager.person import Person

methods = Literal["GET", "PUT", "POST", "PATCH", "DELETE"]
visibility = Literal["private", "internal", "public"]
access_levels = Literal["no_access", "minimal", "guest",
                        "reporter", "developer", "maintainer", "owner"]
access_level: Dict[access_levels, int] = {
    "no_access": 0, "minimal": 5, "guest": 10,
    "reporter": 20, "developer": 30, "maintainer": 40, "owner": 50
}


class Gitlab:
    api_token: str
    base_url: str
    timeout: int
    max_threads: int
    api_version: str = "v4"

    is_timeout: bool = False

    def __init__(self, api_token: str, base_url: Optional[str] = None, timeout: int = 5, max_threads: int = 1) -> None:
        self.api_token = api_token
        self.base_url = base_url or "https://gitlab.com"
        self.timeout = timeout
        self.max_threads = max_threads

    def project(self, id_: int) -> 'Project':
        from .project import Project
        return Project(self, id_)

    def group(self, id_: int) -> 'Group':
        from .group import Group
        return Group(self, id_)

    def send_request(
        self, method: methods, url: str, data: Any = None, params_: Optional[Dict[str, str]] = None, max_tries: int = 5
    ) -> requests.Response:
        try:
            if self.is_timeout:
                time.sleep(5)
            params = params_.copy() if params_ else {}
            params["access_token"] = self.api_token
            ret: requests.Response
            if method == "GET":
                ret = requests.get(
                    f"{self.base_url}/api/{self.api_version}/{url}",
                    params=params
                )
            elif method == "PUT":
                ret = requests.put(
                    f"{self.base_url}/api/{self.api_version}/{url}",
                    json=data, params=params
                )
            elif method == "POST":
                ret = requests.post(
                    f"{self.base_url}/api/{self.api_version}/{url}",
                    json=data,
                    params=params
                )
            elif method == "PATCH":
                ret = requests.patch(
                    f"{self.base_url}/api/{self.api_version}/{url}",
                    json=data,
                    params=params
                )
            elif method == "DELETE":
                ret = requests.delete(
                    f"{self.base_url}/api/{self.api_version}/{url}",
                    params=params
                )
            else:
                raise NotImplementedError
            if ret.status_code == 429:
                # print("\nTimeout!")
                self.is_timeout = True
                time.sleep(self.timeout)
                return self.send_request(method, url, data, params_)
            self.is_timeout = False
            return ret
        except requests.exceptions.RequestException as ex:
            if max_tries == 1:
                raise ex
            time.sleep(5)
            return self.send_request(method, url, data, params_, max_tries - 1)

    def send_requests(
        self, requests_: Iterable[Tuple[methods, str, Any, Optional[Dict[str, str]]]]
    ) -> Iterable[requests.Response]:
        def _send_wrapper(args: Tuple[methods, str, Any, Optional[Dict[str, str]]]) -> requests.Response:
            return self.send_request(*args)

        executor = ThreadPoolExecutor(max_workers=self.max_threads)
        return executor.map(_send_wrapper, requests_)

    def project_members(self, projects: List[int]) -> 'Dict[int, List[User]]':
        def _send_wrapper(project: int) -> 'Tuple[int, List[User]]':
            return project, self.project(project).members()

        ret: 'Dict[int, List[User]]' = {}
        executor = ThreadPoolExecutor(max_workers=self.max_threads)

        for i, users in enumerate(executor.map(_send_wrapper, projects)):
            print(f"Getting project members {(i+1):04d}/{len(projects):04d}", end="\r")
            ret[users[0]] = users[1]
        print()
        return ret

    def search(self, search: str, scope: Literal["projects", "users"]) -> 'List[Dict[str, str | int]]':
        if scope == "projects":
            resp = self.send_request("GET", "projects", params_={"search": search, "order_by": "name", "sort": "asc"})
        else:
            resp = self.send_request("GET", "search", params_={"scope": scope, "search": search})
        return resp.json()

    def search_persons(self, users: 'List[Person]') -> None:
        def _send_wrapper(user: 'Person') -> 'Optional[Tuple[Person, List[Dict[str, str | int]]]]':
            resp = self.search(user.reverseName, "users")
            resp = [r for r in resp if user.checkGitLabName(str(r.get("name", "")))]
            if len(resp) > 1:
                return user, resp
            if len(resp) == 0:
                resp = self.search(user.name, "users")
                resp = [r for r in resp if user.checkGitLabName(str(r.get("name", "")))]
            if len(resp) > 1:
                return user, resp
            if len(resp) == 0:
                return None
            if user.checkGitLabName(str(resp[0].get("name", ""))):
                user.gitlab_id = str(resp[0].get("id", ""))
                user.gitlab_username = str(resp[0].get("username", ""))
            return None

        executor = ThreadPoolExecutor(max_workers=self.max_threads)
        ret: 'List[Tuple[Person, List[Dict[str, str | int]]]]' = []

        for i, rets in enumerate(executor.map(_send_wrapper, users)):
            if rets:
                ret.append(rets)
            print(f"Getting gitlab id {(i+1):04d}/{len(users):04d}", end="\r")
        print()

        for user, search in ret:
            print(f"\nSearch for Person({user.name}) returned multiple options:")
            print("  0: None")
            for i, ser in enumerate(search):
                print(f"  {i + 1}: id: {ser['id']}, name: {ser['name']}, username: {ser['username']}")
            choice = input("--> ")
            try:
                c = int(choice)
                if 0 <= c - 1 < len(search):
                    user.gitlab_id = str(search[c - 1].get("id", ""))
                    user.gitlab_username = str(search[c - 1].get("username", ""))
            except ValueError:
                continue


__all__ = ["Gitlab"]



from . import Gitlab, methods
from typing import Any, Optional, Dict, List, Iterable, Tuple
import requests
import logging

log = logging.getLogger(__name__)


class Group:
    gitlab: Gitlab
    id: int

    def __init__(self, gitlab: Gitlab, id_: int) -> None:
        self.gitlab = gitlab
        self.id = id_

    def send_request(
        self, method: methods, url: str, data: Any = None, params_: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        return self.gitlab.send_request(method, f"groups/{self.id}/{url}", data, params_)

    def send_requests(
        self, requests: Iterable[Tuple[methods, str, Any, Optional[Dict[str, str]]]]
    ) -> Iterable[requests.Response]:
        return self.gitlab.send_requests((x[0], f"groups/{self.id}/{x[1]}", x[2], x[3]) for x in requests)

    def get_info(self) -> 'Optional[Dict[str, int | str]]':
        resp = self.send_request("GET", "", params_={"with_projects": "false"})
        if resp.ok:
            return resp.json()
        else:
            log.debug(resp.json())
        return None

    def projects(self) -> List[int]:
        ret: List[int] = []
        resp = self.send_request("GET", "projects", params_={
                                 "per_page": "100", "page": "1"})
        if not resp.ok:
            log.debug(resp.json())
            return []
        ret.extend(x["id"] for x in resp.json())
        total_pages = resp.headers['X-Total-Pages']
        if total_pages:
            for j, resp in enumerate(
                self.send_requests(
                    ("GET", "projects", None, {"per_page": "100", "page": str(i)}) for i in range(2, int(total_pages)+1)
                )
            ):
                print(f"Getting project page {int(j + 2):02d}/" +
                      f"{int(total_pages):02d}", end="\r")
                ret.extend(x["id"] for x in resp.json())
        print()
        return ret



import datetime
import re
import sys
import toml
import logging
import unicodedata
from .gitlab import Gitlab
from typing import List, Optional, Dict
from .person import Person
from .team import Team
from .utils import parse_csv

log = logging.getLogger(__name__)


class ConfigRepo:
    fullconfig_path: str
    database_path: str

    apitoken: str
    config_repo_id: int
    branch: str
    gitlab_base_url: str

    git: Gitlab
    dryrun: bool = False

    fullconfig: 'Optional[FullConfig]' = None
    persons: 'Optional[List[Person]]' = None
    teams: 'Optional[Dict[str, Team]]' = None

    def __init__(
        self, fullconfig_path: str, database_path: str, apitoken: str, config_repo_id: int,
        branch: str, gitlab_base_url: str, timeout: int, max_thread: int
    ) -> None:
        self.fullconfig_path = fullconfig_path
        self.database_path = database_path
        self.apitoken = apitoken
        self.config_repo_id = config_repo_id
        self.branch = branch
        self.gitlab_base_url = gitlab_base_url

        self.git = Gitlab(self.apitoken, self.gitlab_base_url,
                          timeout, max_thread)

    @classmethod
    def from_toml(cls, data: str, timeout: int = 5, max_thread: int = 1) -> 'ConfigRepo':
        try:
            decoded = toml.loads(data)
            return ConfigRepo(
                decoded["ConfigRepo"]["ConfigPath"],
                decoded["ConfigRepo"]["DatabasePath"],
                decoded["ConfigRepo"]["APIToken"],
                int(decoded["ConfigRepo"]["ConfigRepoID"]),
                decoded["ConfigRepo"]["Branch"],
                decoded["ConfigRepo"]["GitlabBaseURL"],
                timeout,
                max_thread
            )
        except Exception:
            log.critical("Malformed Config File!")
            sys.exit(1)

    def ensure_fullconfig(self, file_name: Optional[str] = None) -> None:
        if self.fullconfig and file_name is None:
            return
        data: str
        try:
            if not file_name:
                log.info(f"Getting {self.fullconfig_path} from configrepo")
                data = self.git.project(self.config_repo_id).get_file(
                    self.fullconfig_path, self.branch)
            else:
                log.debug(f"Read fullconfig from {file_name}")
                with open(file_name, "r") as f:
                    data = f.read()
            self.fullconfig = FullConfig.from_toml(data)
            log.debug(repr(self.fullconfig))
        except Exception:
            log.critical("Malformed FullConfig File!")
            sys.exit(1)

    def ensure_database(self, file_name: Optional[str] = None) -> None:
        if self.persons:
            if not self.teams:
                self.teams = {}
            return
        try:
            data: str
            if not file_name:
                log.info(f"Getting {self.database_path} from configrepo")
                data = self.git.project(self.config_repo_id).get_file(
                    self.database_path, self.branch)
            else:
                try:
                    log.debug(f"Read database from {file_name}")
                    with open(file_name, "r") as f:
                        data = f.read()
                except FileNotFoundError:
                    data = ""
            lines = data.splitlines()
            if len(lines):
                self.persons, self.teams = parse_csv(
                    lines[0].split(";"), lines[1:])
            else:
                self.persons = []
                self.teams = {}
            log.info(f"Read {len(self.persons)} students from database.")
            log.info(f"Read {len(self.teams)} teams from database.")
        except Exception:
            log.critical("Malformed Database!")
            sys.exit(1)

    def add_person(self, person: 'Person') -> bool:
        if self.persons is None:
            return False
        for p in self.persons:
            if p == person:
                p.combine(person)
                return False
        self.persons.append(person)
        return True

    def __repr__(self) -> str:
        return (f"ConfigRepo({self.fullconfig_path}, {self.database_path}, " +
                f"{self.apitoken}, {self.config_repo_id}, {self.branch}, " +
                f"{self.gitlab_base_url})")


class Assignment:
    find_team_regex: Optional[List[re.Pattern]] = None
    find_group_regex: Optional[List[re.Pattern]] = None

    name: str
    repository_name: str
    deadline: Optional[datetime.datetime] = None
    upstream_repo_id: int
    upstream_file_tree: Optional[List[str]] = None

    teamsize: int
    same_team_as: str

    def __init__(
        self, name: str, repository_name: str, upstream_repo_id: int, teamsize: int, same_team_as: str,
        deadline: Optional[datetime.datetime] = None, find_team_regex: Optional[List[str]] = None,
        find_group_regex: Optional[List[str]] = None
    ) -> None:
        if find_team_regex:
            self.find_team_regex = [re.compile(x) for x in find_team_regex]
        if find_group_regex:
            self.find_group_regex = [re.compile(x) for x in find_group_regex]

        self.name = name
        self.repository_name = repository_name
        self.upstream_repo_id = upstream_repo_id
        self.deadline = deadline

        self.teamsize = teamsize
        self.same_team_as = same_team_as

    def get_upstream_repo(self, config: ConfigRepo) -> None:
        if self.upstream_repo_id == -1:
            return
        info = config.git.project(self.upstream_repo_id).get_info()
        if not info or info.get("id") != self.upstream_repo_id:
            return
        self.upstream_file_tree = config.git.project(self.upstream_repo_id).tree(recursive=True)

    def filter_by_group(self, person: 'Person') -> str:
        if self.find_group_regex and len(self.find_group_regex):
            group = person.groups
            for reg in self.find_group_regex:
                res = reg.search(group)
                if not res:
                    return ""
                group = res.string[res.start():res.end()]
            return group
        return person.groups

    def get_team_id(self, person: 'Person', all_persons: 'List[Person]') -> 'Optional[int | str]':
        if person.teams.get(self.name):
            return person.teams[self.name].id

        if not self.filter_by_group(person):
            return None

        if self.find_team_regex and len(self.find_team_regex):
            group = person.groups
            for reg in self.find_team_regex:
                res = reg.search(group)
                if not res:
                    return None
                group = res.string[res.start():res.end()]
            try:
                return int(group)
            except ValueError:
                return group

        if self.same_team_as:
            if person.teams.get(self.same_team_as):
                return person.teams[self.same_team_as].id

        max_num = 1
        for pers in all_persons:
            if pers.teams.get(self.name):
                team = pers.teams[self.name]
                if len(team.persons) < self.teamsize:
                    return team.id
                try:
                    max_num = max(max_num, int(team.id) + 1)
                except ValueError:
                    max_num = max_num
        return max_num

    def fill_repo_name(self, person: 'Optional[Person]' = None, team: 'Optional[Team]' = None) -> str:
        return unicodedata.normalize('NFKD', self.repository_name.replace(
            "<Name>", person.name if person else ""
        ).replace(
            "<Team>", (f"{team.id:03d}" if isinstance(team.id, int) else team.id) if team else ""
        ).replace(
            "<Group>", self.filter_by_group(person) if person else ""
        ).replace(
            "<MatriculationNr>", person.mat_nr if person else ""
        ).replace(
            "<Email>", person.email if person else ""
        ).replace(
            "<GitlabID>", person.gitlab_id if person else ""
        ).replace(
            "<GitlabUsername>", person.gitlab_username if person else ""
        ).replace(
            "<Assignment>", self.name
        )).encode('ascii', 'ignore').decode("ascii")

    def __repr__(self) -> str:
        return (
            f"Assignment({self.name}, {self.repository_name}, {self.upstream_repo_id}, " +
            f"{self.teamsize}, {self.same_team_as}, {self.find_team_regex}, " +
            f"{self.find_group_regex})"
        )


class FullConfig:
    apitoken: str
    baseurl: str
    group_id: int
    group_name: str
    files_to_download: List[str]
    exclude_from_download: List[str]

    assignments: Dict[str, Assignment]
    group_namespace: Optional[str] = None

    def __init__(
        self, apitoken: str, baseurl: str, group_id: int, group_name: str, files_to_download: Optional[List[str]],
        exclude_from_download: Optional[List[str]]
    ) -> None:
        self.assignments = {}
        self.apitoken = apitoken
        self.baseurl = baseurl
        self.group_id = group_id
        self.group_name = group_name
        self.files_to_download = files_to_download or ["*"]
        self.exclude_from_download = exclude_from_download or []

    def get_group_info(self, config: ConfigRepo) -> None:
        if self.group_id == -1:
            return
        info = config.git.group(self.group_id).get_info()
        if not info or info.get("id") != self.group_id:
            return
        self.group_namespace = str(info.get("full_path"))

    @classmethod
    def from_toml(cls, data: str) -> 'FullConfig':
        decoded = toml.loads(data)
        config = FullConfig(
            decoded["Gitlab"]["APIToken"],
            decoded["Gitlab"]["BaseURL"],
            int(decoded["Gitlab"]["GroupID"]),
            decoded["Gitlab"]["GroupName"],
            decoded["Gitlab"].get("FilesToDownload"),
            decoded["Gitlab"].get("ExcludeFilesFromDownload"),
        )

        for ass, c in decoded.get("Assignments", {}).items():
            config.assignments[ass] = Assignment(
                ass,
                c["RepositoryName"],
                int(c["UpstreamRepositoryID"]),
                int(c["Team"]["TeamSize"]),
                c["Team"]["SameTeamAs"],
                datetime.datetime.fromisoformat(c.get("Deadline")) if c.get("Deadline") else None,
                c.get("FindTeamRegex"),
                c.get("FindGroupRegex"),
            )

        return config

    def __repr__(self) -> str:
        return (f"FullConfig({self.apitoken}, {self.baseurl}, " +
                f"{self.group_id}, {self.group_name}, " +
                f"{self.files_to_download}, {self.exclude_from_download}, " +
                f"{self.assignments!r})"
                )


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} [CONFIG_FILE]")
        sys.exit(1)
    with open(sys.argv[1], "r") as f:
        config = ConfigRepo.from_toml(f.read())
    config.ensure_fullconfig()
    print(config)
    print(config.fullconfig)



import datetime
import time
from . import Gitlab, methods, visibility, access_level
from urllib import parse
from typing import Any, Literal, Optional, Dict, List, TYPE_CHECKING, Tuple
import requests
import base64
import logging
from .user import User

if TYPE_CHECKING:
    from esprepomanager.person import Person

log = logging.getLogger(__name__)


class Project:
    gitlab: Gitlab
    id: int

    def __init__(self, gitlab: Gitlab, id_: int) -> None:
        self.gitlab = gitlab
        self.id = id_

    def send_request(
        self, method: methods, url: str, data: Any = None, params_: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        return self.gitlab.send_request(
            method, f"projects/{self.id}/{url}", data, params_)

    def get_info(self) -> Optional[Dict[str, Any]]:
        resp = self.send_request("GET", "")
        if resp.ok:
            return resp.json()
        else:
            log.debug(resp.json())
        return None

    def delete(self, max_tries: int = 3) -> bool:
        for _ in range(max_tries):
            resp = self.send_request("DELETE", "")
            if resp.ok:
                return True
            else:
                log.debug(resp.json())
        return False

    def get_file(self, path: str, ref: str) -> str:
        resp = self.send_request("GET", "repository/files/" + parse.quote(path, ""), params_={"ref": ref})
        if not resp.ok:
            log.debug(resp.json())
            return ""
        return base64.b64decode(resp.json()["content"]).decode("utf-8")

    def get_file_b(self, path: str, ref: str) -> bytes:
        resp = self.send_request("GET", "repository/files/" + parse.quote(path, ""), params_={"ref": ref})
        if not resp.ok:
            log.debug(resp.json())
            return b""
        return base64.b64decode(resp.json()["content"])

    def update_file(self, path: str, branch: str, content: str) -> bool:
        resp = self.send_request("PUT", "repository/files/" + parse.quote(path, ""), {
            "branch": branch,
            "commit_message": f"Updated {path}",
            "content": content,
            "encoding": "text"
        })
        if not resp.ok:
            log.debug(resp.json())
            return False
        return True

    def tree(
        self, path: Optional[str] = None, recursive: bool = False,
        ref: Optional[str] = None, max_tries: int = 3
    ) -> List[str]:
        params = {
            "recursive": "true" if recursive else "false",
            "pagination": "keyset",
            "per_page": "100",
            "order_by": "path",
            "sort": "asc"
        }
        if path:
            params["path"] = path
        if ref:
            params["ref"] = ref

        link: Optional[str] = None
        ret_list: List[str] = []
        next_page: bool = True
        while next_page:
            for _ in range(max_tries):
                if link:
                    resp = requests.get(link)
                else:
                    resp = self.send_request("GET", "repository/tree", params_=params)
                j_resp = resp.json()
                if resp.ok:
                    ret_list.extend(x["path"] for x in j_resp)
                    if resp.links.get("next"):
                        link = resp.links.get("next")["url"]  # type: ignore
                    else:
                        next_page = False
                    break
                else:
                    log.debug(j_resp)
                    time.sleep(1)
            next_page = False

        return ret_list

    def members(self) -> List[User]:
        resp = self.send_request("GET", "members", params_={"per_page": "100"})
        if not resp.ok:
            log.debug(resp.json())
            return []
        return [User.from_json(x) for x in resp.json()]

    def set_protected_branch(self, branch: str, access: int, allow_force: bool = False) -> bool:
        resp = self.send_request("POST", "protected_branches", params_={
            "name": branch,
            "allow_force_push": "true" if allow_force else "false",
            "push_access_level": str(access),
            "merge_access_level": str(access)
        })
        if not resp.ok:
            log.debug(resp.json())
        return resp.ok

    def unprotect_branch(self, branch: str) -> bool:
        resp = self.send_request("DELETE", f"protected_branches/{branch}")
        if not resp.ok:
            log.debug(resp.json())
        return resp.ok

    def fork(
        self, name: str, namespace_id: int, tree: Optional[List[str]],
        visibility_: visibility = "private", members: 'Optional[List[Person]]' = None,
        max_tries: int = 3, branch: str = "main"
    ) -> 'Literal[False] | Tuple[str, int]':
        if self.id == -1:
            return self.create(name, namespace_id, members, max_tries)

        for _ in range(max_tries):
            resp = self.send_request(
                "POST", "fork", params_={
                    "name": name, "path": name, "namespace_id": str(namespace_id), "visibility": visibility_
                }
            )
            j_resp = resp.json()
            log.debug(j_resp)
            is_ok = resp.ok
            try:
                if j_resp.get('message', ['', '', ''])[2] == 'Path has already been taken':
                    rresp = self.gitlab.search(name, "projects")
                    for proj in rresp:
                        if proj["name"] == name:
                            j_resp["id"] = proj["id"]
                            j_resp["web_url"] = proj["web_url"]
                            is_ok = True
                            break
            except IndexError:
                is_ok = False
            if is_ok:
                pid = int(j_resp["id"])
                web_url = j_resp["web_url"]
                time.sleep(1)
                if not tree or self.gitlab.project(pid).tree(recursive=True) == tree:
                    break
                self.gitlab.project(pid).delete(max_tries)
            time.sleep(1)
        else:
            return False

        self.gitlab.project(pid).unprotect_branch(branch)
        self.gitlab.project(pid).set_protected_branch(branch, access_level["developer"], False)

        if members:
            ret = self.gitlab.project(pid).add_members(members, access_level["developer"], max_tries)
            return (web_url, pid) if ret else False
        return (web_url, pid)

    def add_members(self, persons: 'List[Person]', access_level_: int, max_tries: int = 3) -> bool:
        ids = ",".join(x.gitlab_id for x in persons if x.gitlab_id)
        if not len(ids):
            return True

        for _ in range(max_tries):
            resp = self.send_request(
                "POST", "members", params_={"user_id": ids, "access_level": str(access_level_)}
            )

            j_resp = resp.json()
            if resp.ok or j_resp.get('message') == 'Member already exists':
                return True
            log.debug(j_resp)
            time.sleep(1)
        return False

    def create(
        self, name: str, namespace_id: int, members: 'Optional[List[Person]]' = None,
        max_tries: int = 3
    ) -> 'Literal[False] | Tuple[str, int]':
        for _ in range(max_tries):
            resp = self.gitlab.send_request(
                "POST", "projects", params_={"name": name, "namespace_id": str(namespace_id)}
            )

            if resp.ok:
                break
            log.debug(resp.json())
            time.sleep(1)
        else:
            return False

        j_resp = resp.json()
        pid = int(j_resp["id"])
        web_url = j_resp["web_url"]

        if members:
            ret = self.gitlab.project(pid).add_members(members, access_level["developer"], max_tries)
            return (web_url, pid) if ret else ret

        return (web_url, pid)

    def get_most_recent_commit(
        self, date: Optional[datetime.datetime] = None, ref: Optional[str] = None, max_tries: int = 3
    ) -> str:
        params = {}
        if date:
            params["until"] = date.isoformat()+"Z"
        if ref:
            params["ref_name"] = ref
        for _ in range(max_tries):
            resp = self.send_request("GET", "repository/commits", params_=params)

            if resp.ok:
                break
            log.debug(resp.json())
            time.sleep(1)
        else:
            return ""

        j_resp = resp.json()
        if not j_resp:
            return ""
        return j_resp[0].get("id", "")

    def rerun_last_job(self, max_tries: int = 3) -> 'bool':
        for _ in range(max_tries):
            resp = self.send_request("GET", "jobs")
            if not resp.ok:
                continue
            j = resp.json()
            if len(j) == 0:
                return True
            jid = j[0].get("id", None)
            if not jid:
                continue
            resp1 = self.send_request("POST", f"jobs/{jid}/retry")
            if not resp1.ok:
                continue
            return True
        return False

    def set_gitlab_ci_file(self, filename: str, max_tries: int = 3) -> bool:
        for _ in range(max_tries):
            resp = self.send_request("PUT", "", {"ci_config_path": filename})
            if not resp.ok:
                continue
            return True
        return False



class Repository:
    gitlab_server: str
    baseurl: str
    id: int

    def __init__(self, gitlab_server: str, baseurl: str, id_: int) -> None:
        self.gitlab_server = gitlab_server
        self.baseurl = baseurl
        self.id = id_

    @property
    def weburl(self) -> str:
        return f"https://{self.gitlab_server}/{self.baseurl}"

    @property
    def httprepo(self) -> str:
        return f"https://{self.gitlab_server}/{self.baseurl}.git"

    @property
    def sshrepo(self) -> str:
        return f"git@{self.gitlab_server}:{self.baseurl}.git"

    def __repr__(self) -> str:
        return f"Repository({self.gitlab_server}, {self.baseurl}, {self.id})"
